import os
import subprocess
from StringIO import StringIO
from tempfile import mkdtemp

def get_repo_node(env, repo_name, node):
    try:
        from tracopt.versioncontrol.git.PyGIT import GitError
    except ImportError: ## Pre-1.0 Trac
        from tracext.git.PyGIT import GitError
    from trac.core import TracError
    try:
        repo = env.get_repository(reponame=repo_name)
        return repo.get_node(node)
    except GitError:
        raise TracError("Error reading Git files at %s; check your repository path (for repo %s) and file permissions" % (node, repo_name))


def read_config(fp):
    repos = dict()
    this_repo = None
    info = {}
    for line in fp:
        line = line.strip()
        if line.startswith("repo"):
            if this_repo is not None and len(info) > 0:
                repos[this_repo] = info
            this_repo = line[len("repo"):].strip()
            info = {}
        elif '=' in line:
            perms, users = line.split("=")
            perms = perms.strip().upper()
            users = [i.strip() for i in users.split()]
            for perm in perms:
                if perm in info:
                    info[perm].extend(users)
                else:
                    info[perm] = list(users) ## Copy it!

    if this_repo is not None and len(info) > 0:
        repos[this_repo] = info

    fp.close()
    return repos


def to_string(config):

    def _sort(perms):
        tail = []
        ## Ensure the + goes last
        if '+' in perms:
            perms = perms.replace("+", '')
            tail.append("+")
        perms = sorted(perms)
        perms.extend(tail)
        return ''.join(perms)

    fp = StringIO()
    for repo in sorted(config):
        fp.write("repo\t%s\n" % repo)

        ## Combine permissions
        perms = config[repo]
        transposed = {}
        for perm in perms:
            for user in perms[perm]:
                if user in transposed:
                    transposed[user] += perm
                    transposed[user] = _sort(transposed[user])
                else:
                    transposed[user] = perm


        detransposed = {}
        for user in transposed:
            if transposed[user] in detransposed:
                detransposed[transposed[user]].append(user)
                detransposed[transposed[user]] = sorted(detransposed[transposed[user]])
            else:
                detransposed[transposed[user]] = [user]

        for permset in sorted(detransposed):
            fp.write("\t%s\t=\t%s\n" % (permset, ' '.join(detransposed[permset])))

        fp.write("\n")
    fp.seek(0)
    return fp.read()

def save_file(repo_path, file_path, contents, msg):
    tempdir = mkdtemp()
    subprocess.check_call(['git', 'clone', repo_path, tempdir])
    fp = open(os.path.join(tempdir, file_path), 'w')
    fp.write(contents)
    fp.close()
    subprocess.check_call(['git', 'add', file_path], cwd=tempdir)
    subprocess.check_call(['git', 'commit', '-m', msg], cwd=tempdir)
    subprocess.check_call(['git', 'push'], cwd=tempdir)

def remove_files(repo_path, file_paths, msg):
    tempdir = mkdtemp()
    subprocess.check_call(['git', 'clone', repo_path, tempdir])
    for file_path in file_paths:
        subprocess.check_call(['git', 'rm', file_path], cwd=tempdir)
    subprocess.check_call(['git', 'commit', '-m', msg], cwd=tempdir)
    subprocess.check_call(['git', 'push'], cwd=tempdir)
    



from .perm_manager import GitolitePermissionManager
from .perm_policy import GitolitePermissionPolicy
from .repo_manager import GitoliteRepositoryManager
from .user_manager import GitoliteUserManager



from trac.core import *
from trac.config import Option, BoolOption
from trac.perm import IPermissionPolicy

from trac_gitolite import utils

class GitolitePermissionPolicy(Component):
    implements(IPermissionPolicy)

    gitolite_admin_reponame = Option('trac-gitolite', 'admin_reponame',
                                     default="gitolite-admin")
    default_to_private = BoolOption('trac-gitolite', 'default_private',
                                    default=True,
        doc=("If this flag is set to True, then repositories will be private by default, "
             "causing all permissions to be denied to all users if the repository "
             "is not mentioned in the gitolite conf file."))
    all_includes_anonymous = BoolOption('trac-gitolite', 'all_includes_anonymous',
                                        default=False,
        doc=("If this flag is set to True, then anonymous users will be granted permissions "
             "on repositories that specify ``@all = R``.  By default, the ``@all`` token "
             "is considered to mean all logged-in users only."))
    
    ## Helper methods

    def parent_repository(self, resource):
        while True:
            if resource is None:
                return None
            if resource.realm == 'repository':
                return resource
            resource = resource.parent

    def read_config(self):
        node = utils.get_repo_node(self.env, self.gitolite_admin_reponame,
                                   "conf/gitolite.conf")
        fp = node.get_content()
        return utils.read_config(fp)

    def check_repository_permission(self, action, username, repository, resource, perm):
        repos = self.read_config()

        ## If the repo is not known in the config, we defer to the supersystem's decisions,
        ## unless our configuration says otherwise.
        if repository.id not in repos:
            if self.default_to_private:
                return False
            else:
                return None

        perms = repos[repository.id]

        if username == 'anonymous':
            if self.all_includes_anonymous:
                if '@all' in perms.get('R', []):
                    return True
            ## If the repo is known in the config then we assume that anonymous users can't see it
            ## unless the @all-check in the previous stanza was triggered.
            return False

        if username in perms.get('R', []):
            return True

        if '@all' in perms.get('R', []):
            return True

        ## If the repo is known in the config but the user isn't explicitly granted access there,
        ## then the user does not have access.
        return False


    ## IPermissionPolicy methods
            
    def check_permission(self, action, username, resource, perm):
        ## This policy only covers entities that live within a repository
        ## so we'll decline to state any permission if it's not a repository subresource
        repository = self.parent_repository(resource)
        if repository is None:
            return None

        return self.check_repository_permission(action, username, repository, resource, perm)




import pkg_resources

from trac.admin import IAdminPanelProvider
from trac.core import *
from trac.config import Option, BoolOption
from trac.util.translation import _
from trac.web.chrome import ITemplateProvider
from trac.web.chrome import add_notice, add_warning

from trac_gitolite import utils

class GitoliteUserManager(Component):
    implements(IAdminPanelProvider, ITemplateProvider)

    gitolite_admin_reponame = Option('trac-gitolite', 'admin_reponame',
                                     default="gitolite-admin")
    gitolite_admin_ssh_path = Option('trac-gitolite', 'admin_ssh_path',
                                     default="git@localhost:gitolite-admin.git")

    def get_users(self):
        node = utils.get_repo_node(self.env, self.gitolite_admin_reponame, 
                                   "keydir")
        assert node.isdir, "Node %s at /keydir/ is not a directory" % node
        for child in node.get_entries():
            name = child.get_name()
            assert name.endswith(".pub"), "Node %s" % name
            yield name[:-4]

    ## IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'VERSIONCONTROL_ADMIN' in req.perm:
            yield ('versioncontrol', _('Version Control'), 'users', 
                   _('Users'))

    def render_admin_panel(self, req, category, page, path_info):
        req.perm.require('VERSIONCONTROL_ADMIN')

        if req.method == 'POST':

            if 'add_user' in req.args:
                username = req.args['name']
                pubkey = req.args['sshkey'].file
                
                utils.save_file(self.gitolite_admin_ssh_path, 'keydir/%s.pub' % username,
                                pubkey.read(), _('Adding new user %s' % username))

                add_notice(req, _('User "%s" has been added.  Visit the repository permissions panel to set up the new user\'s read/write permissions.' % (
                            username)))
                req.redirect(req.href.admin(category, page))

            users = ['keydir/%s.pub' % username for username in req.args.get('remove_user', [])]
            if users:
                utils.remove_files(self.gitolite_admin_ssh_path, users, _('Removing users'))
                add_notice(req, _('The selected users have been removed and no longer have SSH access to your repositories.  Note that if they have Trac accounts, they may still be able to browse the source code through the web.'))
            else:
                add_warning(req, _('No users were selected.'))
            req.redirect(req.href.admin(category, page))

        data = {'users': self.get_users()}
        return 'admin_repository_users.html', data

    # ITemplateProvider methods

    def get_htdocs_dirs(self):
        return []

    def get_templates_dirs(self):
        return [pkg_resources.resource_filename('trac_gitolite', 'templates')]



import getpass
import json
import pkg_resources

from trac.admin import IAdminPanelProvider
from trac.core import *
from trac.config import Option, BoolOption
from trac.util.translation import _
from trac.web.chrome import ITemplateProvider
from trac.web.chrome import add_notice
from trac.web.chrome import add_warning

from trac_gitolite import utils

class GitolitePermissionManager(Component):
    implements(IAdminPanelProvider, ITemplateProvider)

    gitolite_admin_reponame = Option('trac-gitolite', 'admin_reponame',
                                     default="gitolite-admin")
    gitolite_admin_ssh_path = Option('trac-gitolite', 'admin_ssh_path',
                                     default="git@localhost:gitolite-admin.git")
    gitolite_admin_real_reponame = Option('trac-gitolite', 'admin_real_reponame',
                                          default="gitolite-admin")
    gitolite_admin_system_user = Option('trac-gitolite', 'admin_system_user',
                                        default="trac")

    def get_users(self):
        node = utils.get_repo_node(self.env, self.gitolite_admin_reponame, 
                                   "keydir")
        assert node.isdir, "Node %s at /keydir/ is not a directory" % node
        for child in node.get_entries():
            name = child.get_name()
            assert name.endswith(".pub"), "Node %s" % name
            name = name[:-4]
            yield name

    def read_config(self):
        node = utils.get_repo_node(self.env, self.gitolite_admin_reponame,
                                   "conf/gitolite.conf")
        fp = node.get_content()
        return utils.read_config(fp)
    
    ## IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'VERSIONCONTROL_ADMIN' in req.perm:
            yield ('versioncontrol', _('Version Control'), 'permissions', 
                   _('Permissions'))

    def render_admin_panel(self, req, category, page, path_info):
        req.perm.require('VERSIONCONTROL_ADMIN')

        if req.method == 'POST':
            perms = {}
            for setting in req.args:
                try:
                    setting = json.loads(setting)
                except ValueError:
                    continue
                if not isinstance(setting, dict) or 'perm' not in setting or 'user' not in setting or 'repo' not in setting:
                    continue
                repo = setting['repo']; perm = setting['perm']; user = setting['user']
                if repo not in perms:
                    perms[repo] = {}
                if perm not in perms[repo]:
                    perms[repo][perm] = []
                if user not in perms[repo][perm]:
                    perms[repo][perm].append(user)

            gitolite_admin_perms = perms.get(self.gitolite_admin_real_reponame, {})
            if (self.gitolite_admin_system_user not in gitolite_admin_perms.get('R', []) or
                self.gitolite_admin_system_user not in gitolite_admin_perms.get('W', [])):
                add_warning(req, _('Read and write permissions on the gitolite admin repo must not be revoked for user %s -- otherwise this plugin will no longer work!' % self.gitolite_admin_system_user))
                req.redirect(req.href.admin(category, page))

            utils.save_file(self.gitolite_admin_ssh_path, 'conf/gitolite.conf', 
                            utils.to_string(perms),
                            _('Updating repository permissions'))

            add_notice(req, _('The permissions have been updated.'))
            req.redirect(req.href.admin(category, page))

        perms = self.read_config()

        users_listed_in_perms = set()
        flattened_perms = set()

        for p in perms.values():
            for perm in p:
                flattened_perms.add(perm)
                users_listed_in_perms.update(p[perm])
        flattened_perms = list(flattened_perms)
        def sort_perms(perms):
            tail = []
            ## Ensure the + goes last
            if '+' in perms:
                perms.remove("+")
                tail.append("+")
            perms = sorted(perms)
            perms.extend(tail)
            return perms
        flattened_perms = sort_perms(flattened_perms)

        users = sorted(list(set(list(self.get_users() + list(users_listed_in_perms)))))
        data = {'repositories': perms, 'permissions': flattened_perms, 
                'users': users,
                'sort_perms': sort_perms}
        return 'admin_repository_permissions.html', data

    # ITemplateProvider methods

    def get_htdocs_dirs(self):
        return []

    def get_templates_dirs(self):
        return [pkg_resources.resource_filename('trac_gitolite', 'templates')]



import getpass
import pkg_resources

from trac.admin import IAdminPanelProvider
from trac.core import *
from trac.config import Option, BoolOption
from trac.perm import IPermissionRequestor
from trac.util.translation import _
from trac.web.chrome import ITemplateProvider
from trac.web.chrome import add_notice, add_warning

from trac_gitolite import utils

class GitoliteRepositoryManager(Component):
    implements(IPermissionRequestor, IAdminPanelProvider, ITemplateProvider)

    gitolite_admin_reponame = Option('trac-gitolite', 'admin_reponame',
                                     default="gitolite-admin")
    gitolite_admin_ssh_path = Option('trac-gitolite', 'admin_ssh_path',
                                     default="git@localhost:gitolite-admin.git")

    def read_config(self):
        node = utils.get_repo_node(self.env, self.gitolite_admin_reponame,
                                   "conf/gitolite.conf")
        fp = node.get_content()
        return utils.read_config(fp)
    
    ## IPermissionRequestor methods

    def get_permission_actions(self):
        return [('VERSIONCONTROL_ADMIN', ['REPOSITORY_CREATE']),
                'REPOSITORY_CREATE']

    ## IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'REPOSITORY_CREATE' in req.perm:
            yield ('versioncontrol', _('Version Control'), 'gitolite', 
                   _('Gitolite Repositories'))

    def render_admin_panel(self, req, category, page, path_info):
        req.perm.require('REPOSITORY_CREATE')

        if req.method == 'POST':
            repo_name = req.args['name']
            perms = self.read_config()
            if repo_name in perms:
                add_warning(req, _('A repository named %s already exists; maybe you just need to tell Trac about it using the Repositories panel?'))
                req.redirect(req.href.admin(category, page))
            perms[repo_name] = repo_perms = {}
            trac_user = getpass.getuser()
            for perm in ['R', 'W', '+']:
                repo_perms[perm] = [trac_user]
            utils.save_file(self.gitolite_admin_ssh_path, 'conf/gitolite.conf',
                            utils.to_string(perms),
                            _('Adding new repository %s' % repo_name))
            add_notice(req, _('Repository "%s" has been created.  Now you should give some users permissions on it using the Version Control Permissions panel.' % repo_name))
            req.redirect(req.href.admin(category, page))

        data = {'repos': sorted(self.read_config())}
        return 'admin_repository_gitolite.html', data

    # ITemplateProvider methods

    def get_htdocs_dirs(self):
        return []

    def get_templates_dirs(self):
        return [pkg_resources.resource_filename('trac_gitolite', 'templates')]



from setuptools import setup

try:
    long_description = open("README.rst").read()
except:
    long_description = ''
try:
    long_description += open("CHANGES.txt").read()
except:
    pass

setup(name='trac-GitolitePlugin',
      version='0.3',
      description="",
      long_description=long_description,
      packages=['trac_gitolite'],
      package_data={'trac_gitolite': ['templates/*.html']},
      author='Ethan Jucovy',
      author_email='ejucovy@gmail.com',
      url="http://trac-hacks.org/wiki/GitolitePlugin",
      license='BSD',
      entry_points = {'trac.plugins': ['trac_gitolite = trac_gitolite']})
